# #!/usr/bin/env python
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
# File   : afwprofile.py
#
# Copyright(c) 2023-2025 Farzad Safa (farzad.safa@xcofdk.de)
# This software is distributed under the MIT License (http://opensource.org/licenses/MIT).
# ------------------------------------------------------------------------------

from enum import auto

from _fwadapter                          import rlogif
from _fw.fwssys.fwcore.base.util         import _Util
from _fw.fwssys.fwcore.types.aobject     import _AbsSlotsObject
from _fw.fwssys.fwcore.types.commontypes import _FwIntEnum
from _fw.fwssys.fwcore.types.commontypes import unique
from _fw.fwssys.fwerrh.fwerrorcodes      import _EFwErrorCode

from _fw.fwtdb.fwtdbengine import _EFwTextID
from _fw.fwtdb.fwtdbengine import _FwTDbEngine

class _AbsFwProfile(_AbsSlotsObject):
    @unique
    class _EProfileType(_FwIntEnum):
        eTimer  = 0
        eTask   = auto()
        eThread = auto()

    @unique
    class _EValidationStatus(_FwIntEnum):
        eNone    = 0
        eInvalid = auto()
        eValid   = auto()
        eFrozen  = auto()

    class _ProfileAttributeHandler:
        __slots__ = [ '__h' , '__n' , '__an' ]

        def __init__(self, attrName_, handler_, optionaAttrsNames_ =None):
            self.__h  = handler_
            self.__n  = attrName_
            self.__an = optionaAttrsNames_

        @property
        def handler(self):
            return self.__h

        @property
        def attrName(self):
            return self.__n

        @property
        def optAttrsNames(self):
            return self.__an

    __dictProfileHandlersList = {}
    __slots__ = [ '__pt' , '__ps' , '__aa' ]

    _ATTR_KEY_ARGS                         = _FwTDbEngine.GetText(_EFwTextID.eAbsFwP_AN_ATTR_KEY_ARGS)
    _ATTR_KEY_AUTO_START_ENCLOSED_PYTHREAD = _FwTDbEngine.GetText(_EFwTextID.eAbsFwP_AN_ATTR_KEY_AUTO_START_ENCLOSED_PYTHREAD)
    _ATTR_KEY_ENCLOSED_PYTHREAD            = _FwTDbEngine.GetText(_EFwTextID.eAbsFwP_AN_ATTR_KEY_ENCLOSED_PYTHREAD)
    _ATTR_KEY_KWARGS                       = _FwTDbEngine.GetText(_EFwTextID.eAbsFwP_AN_ATTR_KEY_KWARGS)
    _ATTR_KEY_TASK_NAME                    = _FwTDbEngine.GetText(_EFwTextID.eAbsFwP_AN_ATTR_KEY_TASK_NAME)
    _ATTR_KEY_TASK_RIGHTS                  = _FwTDbEngine.GetText(_EFwTextID.eAbsFwP_AN_ATTR_KEY_TASK_RIGHTS)
    _ATTR_KEY_TASK_ID                      = _FwTDbEngine.GetText(_EFwTextID.eAbsFwP_AN_ATTR_KEY_TASK_ID)

    def __init__(self, ptype_ : _FwIntEnum, profileAttrs_ : dict =None):
        super().__init__()

        self.__pt = None
        self.__aa = {}
        self.__ps = _AbsFwProfile._EValidationStatus.eNone

        if not isinstance(ptype_, _AbsFwProfile._EProfileType):
            rlogif._LogOEC(True, _EFwErrorCode.FE_00416)
        else:
            self.__pt = ptype_
            self._Validate(profileAttrs_)

    @property
    def isValid(self):
        return (self.__ps is not None) and (self.__ps.value >= _AbsFwProfile._EValidationStatus.eValid.value)

    @property
    def isFrozen(self):
        return (self.__ps is not None) and (self.__ps == _AbsFwProfile._EValidationStatus.eFrozen)

    @property
    def isTimerProfile(self):
        return (self.__pt is not None) and self.__pt == _AbsFwProfile._EProfileType.eTimer

    @property
    def isTaskProfile(self):
        return (self.__pt is not None) and self.__pt == _AbsFwProfile._EProfileType.eTask

    @property
    def isThreadProfile(self):
        return (self.__pt is not None) and self.__pt == _AbsFwProfile._EProfileType.eThread

    @property
    def isDrivingXTaskTaskProfile(self):
        return self._isDrivingXTaskTaskProfile

    @property
    def profileStatus(self):
        return self.__ps

    @profileStatus.setter
    def profileStatus(self, val_):
        self.__ps = val_

    @property
    def profileAttributes(self):
        return self.__aa

    @profileAttributes.setter
    def profileAttributes(self, val_):
        self.__aa = val_

    @property
    def internalQueue(self):
        return None

    @property
    def externalQueue(self):
        return None

    def Freeze(self, *args_, **kwargs_):
        if not self.isValid:
            rlogif._LogOEC(True, _EFwErrorCode.FE_00417)
            return False
        if not self.isFrozen:
            self.__ps = _AbsFwProfile._EValidationStatus.eFrozen
        return True

    @staticmethod
    def _IsAutoGeneratedTaskNameEnabled():
        return True

    @staticmethod
    def _SetProfileHandlersList(className_, profileHandlersList_ : list):
        if not _Util.IsInstance(profileHandlersList_, list, bThrowx_=True):
            return False

        _phlist =  _AbsFwProfile.__GetProfileHandlersList(className_)
        if _phlist is not None:
            rlogif._LogOEC(True, _EFwErrorCode.FE_00826)
            return False

        _AbsFwProfile.__dictProfileHandlersList[className_] = profileHandlersList_
        return True

    @staticmethod
    def _CheckMutuallyExclusiveAttrs(dictAttrs_ : dict, mutuallyExclusiveAttrs_ : list):
        if len(dictAttrs_) == 0 or len(mutuallyExclusiveAttrs_) == 0:
            res = True
        else:
            _numMEA = 0
            for _ee in mutuallyExclusiveAttrs_:
                if _ee in dictAttrs_:
                    _numMEA += 1
            res = _numMEA < 2
        if not res:
            rlogif._LogOEC(True, _EFwErrorCode.FE_00418)
        return res

    @property
    def _isDrivingXTaskTaskProfile(self):
        return False

    def _ToString(self):
        return type(self).__name__

    def _CleanUp(self):
        if self.__aa is not None:
            self.__aa.clear()

            self.__aa = None
            self.__ps = None
            self.__pt = None

    def _GetProfileHandlersList(self):
        return _AbsFwProfile.__GetProfileHandlersList(self.__class__.__name__)

    def _GetProfileAttr(self, attrName_, ignoreStatus_ =False):
        res = None
        if self.isValid or ignoreStatus_:
            if attrName_ in self.__aa:
                res = self.__aa[attrName_]
        return res

    def _Validate(self, dictAttrs_ : dict):
        if self._GetProfileHandlersList() is None:
            self.__ps = _AbsFwProfile._EValidationStatus.eInvalid
            rlogif._LogOEC(True, _EFwErrorCode.FE_00419)
            return
        elif self.__ps != _AbsFwProfile._EValidationStatus.eNone:
            self.__ps = _AbsFwProfile._EValidationStatus.eInvalid
            rlogif._LogOEC(True, _EFwErrorCode.FE_00827)
            return
        elif (dictAttrs_ is None) or (len(dictAttrs_) == 0):
            return

        for pah in self._GetProfileHandlersList():
            if pah.attrName not in dictAttrs_:
                continue

            _arg = dictAttrs_[pah.attrName]
            _hh = pah.handler

            _optArgs = None
            if pah.optAttrsNames is not None:
                _optArgs = list()
                for _kk in pah.optAttrsNames:
                    if _kk not in dictAttrs_:
                        _vv = None
                    else:
                        _vv = dictAttrs_[_kk]
                    _optArgs.append(_vv)
                if len(_optArgs) == 0:
                    _optArgs = None

            if _optArgs is not None:
                _hh(self, _arg, *_optArgs)
            else:
                _hh(self, _arg)

    @staticmethod
    def __GetProfileHandlersList(className_):
        res = None
        if className_ in _AbsFwProfile.__dictProfileHandlersList:
            res = _AbsFwProfile.__dictProfileHandlersList[className_]
        return res
